import logging
import asyncio
import datetime
import pytest
import wolf_ism8 as wolf


@pytest.mark.asyncio
async def test_async_setup_server(tst_ism8: wolf.Ism8, _LOGGER):
    _eventloop = asyncio.get_running_loop()
    task1 = _eventloop.create_task(
        _eventloop.create_server(tst_ism8.factory, port=12004)
    )
    _LOGGER.debug("Setup Server")
    _server = await task1
    _LOGGER.debug(f"Waiting for ISM8 connection on {_server.sockets[0].getsockname()}")
    assert _server is not None
    return _server


@pytest.mark.asyncio
async def test_version(_LOGGER):
    """
    checks version reporting
    """
    _LOGGER.debug(f"Lib version {wolf.__version__}")
    assert wolf.__version__ is not None


@pytest.mark.asyncio
async def test_write_boolean_DP(tst_ism8: wolf.Ism8, _LOGGER):
    """
    72:  ('MK1', 'Mischer Zeitprogramm 1', 'DPT_Switch', True)
    """
    _LOGGER.debug("trying to activate MK1 Zeitprogramm Nbr 1")
    encoded_value = tst_ism8.encode_datapoint(72, 1)
    update_msg = tst_ism8.build_message(72, encoded_value)
    assert update_msg is not None


@pytest.mark.asyncio
async def test_write_float_DP(tst_ism8: wolf.Ism8, _LOGGER):
    """
    56: ("DKW", "Warmwassersolltemperatur", "DPT_Value_Temp", True),
    """
    _LOGGER.debug("trying to change warmwasserSollTemp to 51.4")
    tst_ism8.send_dp_value(56, 51.8)
    encoded_value = tst_ism8.encode_datapoint(51.8, 56)
    assert encoded_value is not None
    update_msg = tst_ism8.build_message(56, encoded_value)
    assert update_msg is not None
    await asyncio.sleep(1)


@pytest.mark.asyncio
async def test_network_decoding(tst_ism8: wolf.Ism8, _LOGGER, caplog):
    # decoding tests
    caplog.set_level(logging.DEBUG)
    _LOGGER.debug("trying to decode network msg from github. should fail.")
    test_bytes = bytearray(
        b"\x06\x20\xf0\x80\x00\x14\x04\x00\x00\x00\xf0\x06\x00\xb2\x00\x01\x00\xb2\x03\x00"
    )
    assert tst_ism8.data_received(test_bytes) is True
    # network message should not yield datapoint in dictionary
    assert 178 not in tst_ism8._dp_values.keys()

    _LOGGER.debug("trying to decode valid network msg")
    test_bytes = bytearray(
        b"\x06\x20\xf0\x80\x00\x16\x04\x00\x00\x00\xf0\x06\x00\xb2\x00\x01\x00\xb2\x03\x02\x02\x62"
    )
    assert tst_ism8.data_received(test_bytes) is True
    assert 178 in tst_ism8._dp_values.keys()
    assert tst_ism8._dp_values[178] == pytest.approx(6.1)

    _LOGGER.debug("trying to decode compound network msg1")
    test_bytes = bytearray(
        b"\x06\x20\xf0\x80\x00\x1c\x04\x00\x00\x00\xf0\x06\x00\xb2\x00\x02\x00\xb2\x03\x02\x02\x62\x00\xb3\x03\x02\x02\x63"
    )
    assert tst_ism8.data_received(test_bytes) is True
    assert 178 in tst_ism8._dp_values.keys()
    assert tst_ism8._dp_values[178] == pytest.approx(6.1)
    assert 179 in tst_ism8._dp_values.keys()
    assert tst_ism8._dp_values[178] == pytest.approx(6.1)

    _LOGGER.debug("trying to decode real compound network msg with 5x same message")
    test_bytes = bytearray(
        b"\x06\x20\xf0\x80\x00\x15\x04\x00\x00\x00\xf0\x06\x00\x75\x00\x01\x00\x75\x03\x01\x00\x06\x20\xf0\x80\x00\x15\x04\x00\x00\x00\xf0\x06\x00\x75\x00\x01\x00\x75\x03\x01\x00\x06\x20\xf0\x80\x00\x15\x04\x00\x00\x00\xf0\x06\x00\x75\x00\x01\x00\x75\x03\x01\x00\x06\x20\xf0\x80\x00\x15\x04\x00\x00\x00\xf0\x06\x00\x75\x00\x01\x00\x75\x03\x01\x00\x06\x20\xf0\x80\x00\x15\x04\x00\x00\x00\xf0\x06\x00\x75\x00\x01\x00\x75\x03\x01\x00"
    )
    assert tst_ism8.data_received(test_bytes) is True
    assert 178 in tst_ism8._dp_values.keys()
    assert tst_ism8._dp_values[178] == pytest.approx(6.1)
    assert 179 in tst_ism8._dp_values.keys()
    assert tst_ism8._dp_values[178] == pytest.approx(6.1)

    _LOGGER.debug("trying to decode 1byte valid network msg")
    test_bytes = bytearray(
        b"\x06\x20\xf0\x80\x00\x15\x04\x00\x00\x00\xf0\x06\x00\xb2\x00\x01\x00\xb2\x03\x01\x0a"
    )
    assert tst_ism8.data_received(test_bytes) is True
    assert 178 in tst_ism8._dp_values.keys()
    assert tst_ism8._dp_values[178] == pytest.approx(0.1)


@pytest.mark.asyncio
async def test_date_implementation(tst_ism8: wolf.Ism8, _LOGGER):
    """test of date implementation"""
    _LOGGER.debug("trying to encode date 2024-05-30")
    assert tst_ism8.encode_datapoint(datetime.date(2024, 5, 30), 154) is not None

    # return if value is out of range
    assert wolf.validate_dp_range(154, datetime.date(2024, 5, 30)) is True

    # return if value is out of range
    _LOGGER.debug("trying to encode date 2100-05-30. should be out of range")
    assert wolf.validate_dp_range(154, datetime.date(2100, 5, 30)) is False

    # decoding tests
    _LOGGER.debug("trying to decode date 2007-06-04")
    test_bytes = bytearray(b"\x04\x06\x07")
    tst_ism8.decode_datapoint(159, test_bytes)
    assert tst_ism8._dp_values[159] == datetime.date(2007, 6, 4)

    _LOGGER.debug("trying to decode date 2032-12-20")
    test_bytes = bytearray(b"\x14\x0C\x20")
    # 20.12.2016
    tst_ism8.decode_datapoint(159, test_bytes)
    assert tst_ism8._dp_values[159] == datetime.date(2032, 12, 20)

    _LOGGER.debug("trying to decode date 2048-12-48 (!) should fail")
    test_bytes = bytearray(b"\x30\x0C\x30")
    # 20.12.2016
    assert tst_ism8.decode_datapoint(159, test_bytes) is None

    _LOGGER.debug("trying to decode date from github log 1")
    test_bytes = bytearray(b"\x15\x05\x18")
    tst_ism8.decode_datapoint(155, test_bytes)

    _LOGGER.debug("encode/decode roundtrip")
    test_bytes = tst_ism8.encode_datapoint(datetime.date(2024, 5, 30), 154)
    assert test_bytes
    tst_ism8.decode_datapoint(155, test_bytes)
    assert tst_ism8._dp_values[155] == datetime.date(2024, 5, 30)


@pytest.mark.asyncio
async def test_time_of_day_implementation(tst_ism8: wolf.Ism8):
    """ """
    print("trying to encode timeofday 12:00")
    assert tst_ism8.encode_datapoint(datetime.time(hour=12, minute=0), 161) is not None

    # decoding tests
    print("trying to decode time 22:06:07")
    test_bytes = bytearray(b"\x16\x06\x07")
    tst_ism8.decode_datapoint(161, test_bytes)

    print("trying to decode time 00:00:00")
    test_bytes = bytearray(b"\x00\x00\x00")
    tst_ism8.decode_datapoint(161, test_bytes)
    assert tst_ism8._dp_values[161] == datetime.time(hour=0, minute=0)

    print("trying to decode time from github log 1")
    test_bytes = bytearray(b"\x0d\x38\x00")
    tst_ism8.decode_datapoint(156, test_bytes)
    assert tst_ism8._dp_values[156] == datetime.time(hour=13, minute=56)

    print("trying to decode time from github log 1")
    test_bytes = bytearray(b"\x10\x38\x00")
    tst_ism8.decode_datapoint(157, test_bytes)
    assert tst_ism8._dp_values[157] == datetime.time(hour=16, minute=56)

    print("trying to decode time 48:12:116 (!) should fail, but datetime is robust")
    test_bytes = bytearray(b"\x30\x0C\x60")
    tst_ism8.decode_datapoint(161, test_bytes)

    print("encode/decode roundtrip")
    test_bytes = tst_ism8.encode_datapoint(datetime.time(hour=15, minute=38), 156)
    assert test_bytes is not None
    tst_ism8.decode_datapoint(156, test_bytes)
    assert tst_ism8._dp_values[156] == datetime.time(hour=15, minute=38)


@pytest.mark.asyncio
async def test_scaling_implementation(tst_ism8: wolf.Ism8):
    """
    test scaling implementation
    """
    assert wolf.encode_scaling(100) == b"\xff"
    assert wolf.encode_scaling(0) == b"\x00"


@pytest.mark.asyncio
async def test_write_HVACMode57(tst_ism8: wolf.Ism8, _LOGGER):
    """
    57 Programmwahl Heizkreis DPT_HVACMode Out / In
    """
    _LOGGER.debug("trying to change HVAC modes")
    # not in range
    assert wolf.validate_dp_range(57, "Comfort") is False
    # ok
    assert wolf.validate_dp_range(57, "Automatikbetrieb") is True
    assert wolf.validate_dp_range(57, "Heizbetrieb") is True
    assert wolf.validate_dp_range(57, "Standby") is True
    assert wolf.validate_dp_range(57, "Automatikbetrieb kühlen") is True


@pytest.mark.asyncio
async def test_write_HVACMode149(tst_ism8: wolf.Ism8, _LOGGER):
    """
    149 Programmwahl Heizkreis DPT_HVACMode_CWL Out / In
    """
    _LOGGER.debug("test HVAC modes for CWL and direct heating")
    # not in range
    assert wolf.validate_dp_range(149, "Comfort") is False
    # ok
    assert wolf.validate_dp_range(149, "Automatikbetrieb") is True
    assert wolf.validate_dp_range(149, "Heizbetrieb") is False
    assert wolf.validate_dp_range(149, "Standby") is True
    assert wolf.validate_dp_range(149, "Feuchteschutz") is True
    # not in range
    assert wolf.validate_dp_range(70, "Comfort") is False
    # ok
    assert wolf.validate_dp_range(70, "Automatikbetrieb") is True
    assert wolf.validate_dp_range(70, "Heizbetrieb") is True
    assert wolf.validate_dp_range(70, "Standby") is True
    assert wolf.validate_dp_range(70, "Feuchteschutz") is False


@pytest.mark.asyncio
async def test_write_DHWMode(tst_ism8: wolf.Ism8):
    """
    58 Programmwahl Warmwasser DPT_DHWMode Out / In
    """
    # not in range
    assert wolf.validate_dp_range(58, "GibtsNicht") is False
    assert wolf.validate_dp_range(58, "Automatikbetrieb") is True
    assert wolf.validate_dp_range(58, "LegioProtect") is False


@pytest.mark.asyncio
async def test_HVACCONTRMode(tst_ism8: wolf.Ism8):
    """
    177 Betriebsart DPT_HVACContrMode
    """
    tst_ism8.decode_datapoint(177, bytearray(b"\x01"))
    assert tst_ism8._dp_values[177] == "Heizbetrieb"
    tst_ism8.decode_datapoint(177, bytearray(b"\x06"))
    assert tst_ism8._dp_values[177] == "Standby"
    tst_ism8.decode_datapoint(177, bytearray(b"\x07"))
    assert tst_ism8._dp_values[177] == "Test"
    tst_ism8.decode_datapoint(177, bytearray(b"\x08"))
    assert tst_ism8._dp_values[177] == "Emergency Heat"
    tst_ism8.decode_datapoint(177, bytearray(b"\x09"))
    assert tst_ism8._dp_values[177] == "Fan Only"

    assert tst_ism8.encode_datapoint("GibtsNicht", 177) is None
    assert tst_ism8.encode_datapoint("Auto", 177) == b"\x00"
    assert tst_ism8.encode_datapoint("Frostschutz", 177) == b"\x0b"


@pytest.fixture(scope="module")
def tst_ism8():
    return wolf.Ism8()


@pytest.fixture(scope="module")
def _LOGGER():
    return logging.getLogger(__name__)



# pylint: disable-msg=W1203
"""contains encode/decode helper functions for ISM8 datapoints"""
import logging
import datetime
from .ism8_constants import (
    DP_VALUES_ALLOWED,
    DATAPOINTS,
    DT_PYTHONTYPE,
    IX_RW_FLAG,
    IX_TYPE,
    DATATYPES,
)

log = logging.getLogger(__name__)


def decode_dict(mode_number: int, mode_dic: dict) -> str | None:
    """returns a human readable string from the API-encoded mode_number"""
    if mode_number in mode_dic.keys():
        return mode_dic[mode_number]
    else:
        log.error(f"mode number {mode_number} not implemented:")
        return None


def encode_dict(mode: str, mode_dic: dict) -> bytearray | None:
    """encodes a string into corresponding ISM-Mode numbers"""
    entry_list = [item[0] for item in mode_dic.items() if item[1] == mode]
    if not entry_list:
        log.error(f"error encoding {mode}")
        log.error(f"available modes: {mode_dic.items()}")
        return None
    if len(entry_list) == 1:
        # the bytearray-constructor NEEDS a list with one entry!
        # do not cast the mode-number on its own
        return bytearray(entry_list)
    else:
        log.error(f"error encoding mode {mode}, matching not exact ")
        return None


def decode_scaling(value: int) -> float:
    """decodes a scaled value from 0-255 to 0-100"""
    return 100 / 255 * value


def encode_scaling(value: float) -> bytearray:
    """encodes a value from 0-100 to 0-255"""
    return bytearray([round(value / (100 / 255))])


def decode_bool(value: int) -> bool:
    """decodes a boolean value for int data"""
    return bool(value & 0b1)


def encode_bool(value: int) -> bytearray:
    """encodes a boolean value from int data"""
    return bytearray(b"\x01") if bool(value) is True else bytearray(b"\x00")


def decode_int(value: int) -> int:
    """decodes an integer value"""
    return int(value)


def decode_float(value: int) -> float | None:
    """decodes a float value"""
    _sign = (value & 0b1000000000000000) >> 15
    _exponent = (value & 0b0111100000000000) >> 11
    _mantisse = value & 0b0000011111111111
    if _mantisse == 0b0000011111111111:
        # according to WOLF specs, a mantisse with all bits set
        # indicates invalid data
        return None
    if _sign == 1:
        _mantisse = -(~(_mantisse - 1) & 0x07FF)
    decoded_float = float(0.01 * (2**_exponent) * _mantisse)
    return decoded_float


def encode_float(value: float) -> bytearray:
    """encodes a float value to ISM8 format"""
    value = round(value, 2)
    data = [0, 0]
    encoded_float = bytearray()
    _exponent = 0
    _mantisse_calc = round(abs(value) * 100)
    while _mantisse_calc.bit_length() > 11:
        _exponent += 1
        _mantisse_calc = round(_mantisse_calc / 2)
    _mantisse = round(value * 100 / (1 << _exponent))
    if value < 0:
        data[0] |= 0x80
        _mantisse = round((~(_mantisse * -1) + 1) & 0x07FF)
    data[0] |= (_exponent & 0x0F) << 3
    data[0] |= (_mantisse >> 8) & 0x7
    data[1] |= _mantisse & 0xFF
    for byte in data:
        encoded_float.append(byte)
    # log.debug(f"encoded {input} -> {encoded_float.hex(':')}")
    return encoded_float


def decode_date(value: int) -> datetime.date:
    """decodes a date value"""
    year = value & 0b000000000000000001111111
    month = (value & 0b000000000000111100000000) >> 8
    day = (value & 0b000111110000000000000000) >> 16
    return datetime.date(year + 2000, month, day)


def encode_date(value: datetime.date) -> bytearray:
    """encodes a date value"""
    encoded_date = bytearray()
    encoded_date.append(value.day)
    encoded_date.append(value.month)
    encoded_date.append(value.year - 2000)
    log.debug(f"encoded {value} -> {encoded_date.hex(':')}")
    return encoded_date


def decode_time_of_day(value: int) -> datetime.time:
    """decodes a time value"""
    seconds = value & 0b000000000000000000111111
    minutes = (value & 0b000000000011111100000000) >> 8
    hours = (value & 0b000111110000000000000000) >> 16
    return datetime.time(hour=hours, minute=minutes, second=seconds)


def encode_time_of_day(value: datetime.time) -> bytearray:
    """encodes a time value"""
    encoded_time = bytearray()
    encoded_time.append(value.hour)
    encoded_time.append(value.minute)
    encoded_time.append(value.second)
    log.debug(f"encoded {value} -> {encoded_time.hex(':')}")
    return encoded_time


def validate_dp_range(dp_id: int, value) -> bool:
    """checks if value is valid for the datapoint before sending to ISM"""
    # check if dp is R/O
    if not DATAPOINTS[dp_id][IX_RW_FLAG]:
        log.error(f"datapoint {dp_id} is not writable")
        return False

    # check if datatype is as expected
    dp_type = DATAPOINTS[dp_id][IX_TYPE]
    python_datatype = DATATYPES[dp_type][DT_PYTHONTYPE]
    if not isinstance(value, python_datatype):
        log.error(f"DP {dp_id} should be {python_datatype}, but is {type(value)}")
        return False

    # check if value is in allowed range
    if isinstance(value, str):
        if value not in DP_VALUES_ALLOWED[dp_id]:
            log.error(f"value {value} is out of range")
            return False
    else:
        if (value > max(DP_VALUES_ALLOWED[dp_id])) or (
            value < min(DP_VALUES_ALLOWED[dp_id])
        ):
            log.error(f"value {value} is out of range")
            return False
    return True
